import java.util.*;

class Main {
	public static void main(String args[]) {
		roundOffError();
		equalsError();
		cloneError();
		danglingElseError();
		offByOneError();
		loopTerminationError();

		(new Puzzle1()).run();
		(new Puzzle2()).run();
		(new Puzzle3()).run();
		(new Puzzle4()).run();
		(new Puzzle5()).run();
	}
	
	/**
	 * Demonstrates round-off error.
	 *<P>
	 * Adapted from Hostmann, p. 55
	 **/
	static void roundOffError() {
		println("---\nroundOffError:");

		double f = 2e15 + 0.13;
		double g = 2e15 + 0.02;
		println(100*(f-g)); // prints 11?
	}
	
	static void println(String s) {
		System.out.println(s);
	}
	
	static void println(double d) {
		System.out.println(d);
	}
	
	/**
	 * Demonstrates difference between equals() and ==
	 *<P>
	 **/
	static void equalsError() {
		println("---\nequalsError:");

		String s1 = new String("This is a string");
		String s2 = new String("This is a string");
		test("String==", s1 == s2);
		test("String.equals", s1.equals(s2));

		Object x = new Object();
		Object y = new Object();
		test("object==", x == y);
		test("object.equals", x.equals(y));

		String s3 = "This is a string";
		String s4 = "This is a string";
		test("String==", s3 == s4);
		test("String.equals", s3.equals(s4));
	}

	static void test(String name, boolean bool) {
		println(name + ": " + (bool?"true":"false"));
	}

	/**
	 * Forgetting to clone an object before modifying it.
	 *<P>
	 * Adapted from Hostmann, p. 116
	 **/
	static void cloneError() {
		println("---\ncloneError:");

		Date now = new Date();
		Date later = now;
		// Date later = new Date(now.getTime());
		later.setHours(now.getHours() + 1);
		if (now.before(later))
			println("see you later");
		else
			println("see you now");
	}

	/**
	 * The dangling else problem.
	 **/
	static void danglingElseError() {
		println("---\ndanglingElseError:");

		checkEven(-1);
		checkEven(0);
		checkEven(1);
	}

	/**
	 * Illustrates the dangling else problem.
	 * This methods should (but doesn't)
	 * report even and negative numbers,
	 * and ignore odd numbers.
	 *<P>
	 * Adapted from Hostmann, p. 156
	 **/
	static void checkEven(int n) {
		boolean result = true;
		if (n>=0)
			if ((n%2) == 0)
				println(n + " is even");
		else
			println(n + " is negative");
	}

	/**
	 * Off-by-1 errors.
	 *<P>
	 * Adapted from Hostmann, p. 244
	 **/
	static void offByOneError() {
		println("---\noffByOneError:");

		showBinomial(7,3);
	}
	
	static void showBinomial(int n, int k) {
		println("binomial(" + n + "," + k + ") = "
			+ binomial(n,k)
			+ " =? "
			+ factorial(n)/(factorial(k)*factorial(n-k)));
	}

	/**
	 * The binomial coefficient is n!/k!*(n-k)!
	 *<P>
	 * The code below is wrong -- check either by
	 * counting iterations, or by checking
	 * boundary conditions.  There should be 
	 * (but aren't) k iterations.
	 * The first factor should be n/1 (ok) and the last
	 * should be (n-k+1)/k (not ok).
	 **/
	static int binomial(int n, int k) {
		int bc = 1;
		for (int i=1; i<k; i++)
			bc = bc * (n+1-i) / i;
		return bc;
	}

	static int factorial(int n) {
		int result=1;
		for (int i=1; i<=n; i++)
			result = result*i;
		return result;
	}

	/**
	 * Don't use =! to test the end of a range.
	 **/
	static void loopTerminationError() {
		println("---\nloopTerminationError:");

		showFactorial(0);
		showFactorial(5);
		// showFactorial(-1);
	}

	static void showFactorial(int n) {
		println("brokenFactorial(" + n + ") = " + brokenFactorial(n));
	}

	/**
	 * Don't use =! to test the end of a range.
	 * This factorial function won't work for -1.
	 *<P>
	 * Adapted from Horstmann, p. 246
	 **/
	static int brokenFactorial(int n) {
		int result=1;
		for (int i=0; i!=n; i++)
			result = result*(i+1);
		return result;
	}
}
